import glm
from OpenGL.GL import GL_LINES

import pxng


class Grid:
    def __init__(self, width, height):
        program = pxng.ShaderProgram('GridShader')
        program.add_shader(pxng.resource('shaders/line.vert'), pxng.ShaderType.Vertex)
        program.add_shader(pxng.resource('shaders/line.frag'), pxng.ShaderType.Fragment)
        program.compile_and_link()

        program.add_uniform('projection_view', glm.mat4x4)
        program.add_uniform('model', glm.mat4x4)
        program.add_uniform('color', glm.vec4)
        program.add_uniform('resolution', glm.vec2)
        program.add_uniform('dash_size', glm.vec1)
        program.add_uniform('gap_size', glm.vec1)

        self._shader_program = program

        self._vao = pxng.VertexArrayObject(GL_LINES)
        self._vao.attach_buffer(pxng.BufferObject(data_type=glm.vec3))  # vertex buffer

        x = 1
        y = 1
        while x < width:
            self._vao.add_line(glm.vec3(x, 0, 0), glm.vec3(x, height, 0))
            x += 1

        while y < height:
            self._vao.add_line(glm.vec3(0, y, 0), glm.vec3(width, y, 0))
            y += 1

    def draw(self, spaces: pxng.Spaces, size, dash_size, gap_size):
        if self._vao.bind():
            spaces.model.push()
            spaces.model.scale((size, size, 1))
            program = self._shader_program
            program.activate()
            program.set_uniform('projection_view', spaces.projection_view)
            program.set_uniform('model', spaces.model.m)
            program.set_uniform('color', spaces.tint)
            program.set_uniform('resolution', glm.vec2(spaces.width, spaces.height))
            program.set_uniform('dash_size', dash_size)
            program.set_uniform('gap_size', gap_size)

            self._vao.draw()

            spaces.model.pop()



from copy import copy

import glm


class CoordinateSystem:

    def __init__(self):
        self._m = glm.mat4x4(1)
        self._stack = []

    def translate(self, vec3):
        self._m = glm.translate(self._m, vec3)

    def scale(self, vec3):
        self._m = glm.scale(self._m, vec3)

    def rotate(self, angle, vec3):
        self._m = glm.rotate(self._m, angle, vec3)

    def push(self):
        self._stack.insert(0, copy(self._m))

    def load_identity(self):
        self._m = glm.mat4x4(1)

    def pop(self):
        self._m = self._stack.pop(0)

    @property
    def m(self):
        return self._m

    @m.setter
    def m(self, m: glm.mat4x4):
        self._m = m


class Spaces:
    def __init__(self, width, height):
        self._width = width
        self._height = height
        self._projection = CoordinateSystem()
        self._view = CoordinateSystem()
        self._model = CoordinateSystem()
        self._texture = CoordinateSystem()
        self._default_tint = glm.vec4(1)
        self._tint = self._default_tint

    @property
    def projection(self) -> CoordinateSystem:
        return self._projection

    @property
    def view(self) -> CoordinateSystem:
        return self._view

    @property
    def model(self) -> CoordinateSystem:
        return self._model

    @property
    def texture(self) -> CoordinateSystem:
        return self._texture

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    @property
    def projection_view(self) -> glm.mat4x4:
        return self.projection.m * self.view.m

    @property
    def tint(self) -> glm.vec4:
        return self._tint

    @tint.setter
    def tint(self, tint: glm.vec4):
        if tint is None:
            tint = self.default_tint

        if len(tint) == 3:
            tint = glm.vec4(*tint, 1)
        else:
            tint = glm.vec4(*tint)
        self._tint = tint

    @property
    def default_tint(self):
        return self._default_tint

    @default_tint.setter
    def default_tint(self, tint):
        self._default_tint = tint

    def push(self):
        self.projection.push()
        self.view.push()
        self.model.push()
        self.texture.push()

    def pop(self):
        self.projection.pop()
        self.view.pop()
        self.model.pop()
        self.texture.pop()



from enum import Enum
import os.path
from typing import List

import glm
import OpenGL.GL as gl


class ShaderType(Enum):
    Vertex = gl.GL_VERTEX_SHADER
    Fragment = gl.GL_FRAGMENT_SHADER
    Geometry = gl.GL_GEOMETRY_SHADER


class _Shader:
    def __init__(self, path, shader_type: ShaderType):
        self._path = path
        self._shader_type = shader_type
        self._id = None
        self._timestamp = os.path.getmtime(path)
        self._failed_version = None
        self._attached = False

    @property
    def id(self):
        return self._id

    @property
    def last_modified(self):
        return self._timestamp

    @property
    def path(self):
        return self._path

    @property
    def type(self):
        return self._shader_type

    @property
    def has_changed(self):
        return self.last_modified != os.path.getmtime(self.path)

    @property
    def needs_compile(self):
        return self._id is None or self.has_changed

    def compile(self):
        with open(self.path, 'r') as fh:
            source = fh.read()
            self._timestamp = os.path.getmtime(self.path)

        if self._failed_version == self._timestamp:
            return

        if self._id is None:
            self._id = gl.glCreateShader(self._shader_type.value)

        gl.glShaderSource(self._id, source)
        gl.glCompileShader(self._id)

        status = gl.glGetShaderiv(self._id, gl.GL_COMPILE_STATUS, None)
        if status == gl.GL_FALSE:
            log = gl.glGetShaderInfoLog(self._id)

            print(f'An error occurred during compilation of: {self._path}')
            print('-' * 20)
            print(f'{log.decode("utf-8")}')
            print('-' * 20)
            self._failed_version = self._timestamp
        else:
            self._failed_version = None

        return status == gl.GL_TRUE

    def failed_compile(self):
        return self._failed_version is not None

    @property
    def attached(self):
        return self._attached

    @attached.setter
    def attached(self, attached):
        self._attached = attached

    def __repr__(self):
        status = self.failed_compile()
        return f'{self._path} id: {self._id} @ {self._timestamp} - status: {status}'


class ShaderProgram:
    def __init__(self, name, frag_data_location=0, frag_data_name='out_color'):
        self._name = name
        self._shaders: List[_Shader] = []
        self._id = None
        self._frag_data_loc = frag_data_location
        self._frag_data_name = frag_data_name
        self._uniforms = {}

    def add_shader(self, path, shader_type: ShaderType):
        self._shaders.append(_Shader(path, shader_type))

    def add_uniform(self, name, data_type):
        self._uniforms[name] = {
            'loc': self.get_uniform_location(name),
            'type': data_type,
            'fn': self._get_gl_function_for_type(data_type)
        }

    @staticmethod
    def _get_gl_function_for_type(data_type):
        if data_type == glm.mat4x4:
            def fn(loc, value):
                gl.glUniformMatrix4fv(loc, 1, False, glm.value_ptr(value))
            return fn
        elif data_type == glm.dvec4:
            def fn(loc, value):
                gl.glUniform4dv(loc, 1, glm.value_ptr(value))
            return fn
        elif data_type == glm.vec4:
            def fn(loc, value):
                gl.glUniform4fv(loc, 1, glm.value_ptr(value))
            return fn
        elif data_type == glm.vec3:
            def fn(loc, value):
                gl.glUniform3fv(loc, 1, glm.value_ptr(value))
            return fn
        elif data_type == glm.vec2:
            def fn(loc, value):
                gl.glUniform2fv(loc, 1, glm.value_ptr(value))
            return fn
        elif data_type == glm.vec1:
            return gl.glUniform1f
        elif data_type == glm.ivec1:
            return gl.glUniform1i
        elif data_type == glm.uvec1:
            return gl.glUniform1ui
        elif data_type == glm.bvec1:
            return gl.glUniform1ui
        else:
            raise UserWarning(f'Unknown data type: {data_type}')

    def _create(self):
        if self._id is None:
            self._id = gl.glCreateProgram()

        for shader in self._shaders:
            if shader.needs_compile:
                shader.compile()

            if shader.failed_compile():
                # One of the shaders did not compile correctly -> skip linking step
                return False

        return self._link_shader()

    def needs_recompile(self):
        for shader in self._shaders:
            if shader.needs_compile:
                return True
        return False

    def activate(self):
        gl.glUseProgram(self._id)

    def _link_shader(self):
        for shader in self._shaders:
            if shader.attached:
                gl.glDetachShader(self._id, shader.id)
            gl.glAttachShader(self._id, shader.id)
            shader.attached = True

        gl.glBindFragDataLocation(self._id, self._frag_data_loc, self._frag_data_name)

        gl.glLinkProgram(self._id)
        status = gl.glGetProgramiv(self._id, gl.GL_LINK_STATUS, None)
        if status == gl.GL_FALSE:
            log = gl.glGetProgramInfoLog(self._id)
            print(f'An error occurred during linking of: {self._name}')
            print('-' * 20)
            print(f'{log.decode("utf-8")}')
            print('-' * 20)
            return False
        return True

    def get_uniform_location(self, name):
        return gl.glGetUniformLocation(self._id, name)

    def set_uniform(self, name, value):
        u = self._uniforms[name]
        u['fn'](u['loc'], value)

    def compile_and_link(self) -> bool:
        return self._create()



from typing import List

import glm
import pxng

import OpenGL.GL as gl


class VertexArrayObject:
    def __init__(self, primitive):
        self._primitive = primitive
        self._buffers: List[pxng.BufferObject] = []
        self._indices = pxng.BufferObject(data_type=self.index_data_type,
                                          array_type=gl.GL_ELEMENT_ARRAY_BUFFER)
        self._vao = gl.glGenVertexArrays(1)

    def attach_buffer(self, vbo: pxng.BufferObject):
        self._buffers.append(vbo)
        return len(self._buffers) - 1

    def add_quad(self, p1, p2, p3, p4):
        i = self._buffers[0].index
        self._buffers[0].set_value(p1)
        self._buffers[0].set_value(p2)
        self._buffers[0].set_value(p3)
        self._buffers[0].set_value(p4)
        self._indices.set_value(glm.u16vec3(i, i + 1, i + 3))
        self._indices.set_value(glm.u16vec3(i + 1, i + 2, i + 3))

    def add_triangle(self, p1, p2, p3):
        i = self._buffers[0].index
        self._buffers[0].set_value(p1)
        self._buffers[0].set_value(p2)
        self._buffers[0].set_value(p3)
        self._indices.set_value(glm.u16vec3(i, i + 1, i + 2))

    def add_line(self, p1, p2):
        i = self._buffers[0].index
        self._buffers[0].set_value(p1)
        self._buffers[0].set_value(p2)
        self._indices.set_value(glm.u16vec2(i, i + 1))

    def add_point(self, p1):
        i = self._buffers[0].index
        self._buffers[0].set_value(p1)
        self._indices.set_value(glm.u16vec1(i))

    def set_colors(self, *args: glm.vec4, target=1):
        for c in args:
            self._buffers[target].set_value(c)

    def set_texture(self, *args: glm.vec2 or glm.uvec2, target=1):
        for c in args:
            self._buffers[target].set_value(c)

    def create(self):
        gl.glBindVertexArray(self._vao)

        for index, vbo in enumerate(self._buffers):
            vbo.bind(index)

        self._indices.bind(None)

    def reset(self):
        self._indices.reset()
        for vbo in self._buffers:
            vbo.reset()

    def draw(self):
        index_count = len(self._indices) * self.primitive_component_count
        gl.glDrawElements(self._primitive, index_count, gl.GL_UNSIGNED_SHORT, None)

    @property
    def index_data_type(self):
        if self._primitive == gl.GL_TRIANGLES:
            return glm.u16vec3
        elif self._primitive == gl.GL_LINES:
            return glm.u16vec2
        elif self._primitive == gl.GL_POINTS:
            return glm.u16vec1
        else:
            raise UserWarning(f'Unknown primitive type {self._primitive}')

    @property
    def primitive_component_count(self):
        if self._primitive == gl.GL_TRIANGLES:
            return 3
        elif self._primitive == gl.GL_LINES:
            return 2
        elif self._primitive == gl.GL_POINTS:
            return 1
        else:
            raise UserWarning(f'Unknown primitive type {self._primitive}')

    def bind(self):
        gl.glBindVertexArray(self._vao)
        if self._indices.bind(None):
            if any(vbo.changed for vbo in self._buffers):
                self.create()
            return True
        gl.glBindVertexArray(0)
        return False



from dataclasses import dataclass
from typing import Tuple

import glfw

MOUSE_BUTTON_1 = glfw.MOUSE_BUTTON_1
MOUSE_BUTTON_2 = glfw.MOUSE_BUTTON_2
MOUSE_BUTTON_3 = glfw.MOUSE_BUTTON_3
MOUSE_BUTTON_4 = glfw.MOUSE_BUTTON_4
MOUSE_BUTTON_5 = glfw.MOUSE_BUTTON_5
MOUSE_BUTTON_6 = glfw.MOUSE_BUTTON_6
MOUSE_BUTTON_7 = glfw.MOUSE_BUTTON_7
MOUSE_BUTTON_8 = glfw.MOUSE_BUTTON_8
MOUSE_BUTTON_LEFT = glfw.MOUSE_BUTTON_LEFT
MOUSE_BUTTON_RIGHT = glfw.MOUSE_BUTTON_RIGHT
MOUSE_BUTTON_MIDDLE = glfw.MOUSE_BUTTON_MIDDLE


@dataclass
class MouseButtonState:
    pressed: bool = False
    """
    bool: This is true if the mouse button was pressed
    """

    released: bool = False
    """
    bool: This is true when the mouse button is released.
    """

    held: bool = False
    """
    bool: This is true from the moment of pressed until released.
    """

    def __str__(self):
        return f'self.pressed={self.pressed} self.released={self.released} self.held={self.held}'


class Mouse:
    def __init__(self, glfw_window):
        self._buttons = {
            MOUSE_BUTTON_1: MouseButtonState(),
            MOUSE_BUTTON_2: MouseButtonState(),
            MOUSE_BUTTON_3: MouseButtonState(),
            MOUSE_BUTTON_4: MouseButtonState(),
            MOUSE_BUTTON_5: MouseButtonState(),
            MOUSE_BUTTON_6: MouseButtonState(),
            MOUSE_BUTTON_7: MouseButtonState(),
            MOUSE_BUTTON_8: MouseButtonState(),
        }

        self._x = 0
        self._y = 0
        self._dx = 0
        self._dy = 0
        self._hover = False
        self._scroll_dx = 0
        self._scroll_dy = 0

        glfw.set_input_mode(glfw_window, glfw.STICKY_MOUSE_BUTTONS, glfw.TRUE)
        glfw.set_scroll_callback(glfw_window, self._scroll_callback)

    @property
    def hover(self) -> bool:
        """
        Is the mouse cursor hovering over the window?
        Returns
        -------
        bool
        """
        return self._hover

    @property
    def pos(self) -> Tuple[float, float]:
        """
        Returns the current mouse position as a tuple of floats
        Returns
        -------
        tuple of (float, float)
        """
        return self._x, self._y

    @property
    def x(self) -> float:
        """
        Returns the mouse x position
        Returns
        -------
        float
        """
        return self._x

    @property
    def y(self) -> float:
        """
        Returns the mouse y position
        Returns
        -------
        float
        """
        return self._y

    @property
    def dx(self) -> float:
        """
        Returns the mouse delta position in x direction since the last poll
        Returns
        -------
        float
        """
        return self._dx

    @property
    def dy(self) -> float:
        """
        Returns the mouse delta position in y direction since the last poll
        Returns
        -------
        float
        """
        return self._dy

    @property
    def scroll_dx(self) -> float:
        """
        Returns the currently accumulated amount of scrolling in the x direction.
        Beware that the scroll accumulator is reset after this call.
        Returns
        -------
        float
        """
        dx = self._scroll_dx
        return dx

    @property
    def scroll_dy(self) -> float:
        """
        Returns the currently accumulated amount of scrolling in the y direction.
        Beware that the scroll accumulator is reset after this call.
        Returns
        -------
        float
        """
        dy = self._scroll_dy
        self._scroll_dy = 0
        return dy

    @property
    def scroll(self) -> Tuple[float, float]:
        """
        Returns the currently accumulated amount of scrolling in the both directions.
        Beware that the scroll accumulator is reset after this call.
        Returns
        -------
        float
        """
        return self.scroll_dx, self._scroll_dy

    @property
    def button_left(self) -> MouseButtonState:
        """
        Returns the state for the left mouse button.
        Returns
        -------
        MouseButtonState
        """
        return self._buttons[MOUSE_BUTTON_LEFT]

    @property
    def button_right(self) -> MouseButtonState:
        """
        Returns the state for the right mouse button.
        Returns
        -------
        MouseButtonState
        """
        return self._buttons[MOUSE_BUTTON_RIGHT]

    @property
    def button_middle(self) -> MouseButtonState:
        """
        Returns the state for the middle mouse button.
        Returns
        -------
        MouseButtonState
        """
        return self._buttons[MOUSE_BUTTON_MIDDLE]

    def button_state(self, button_id) -> MouseButtonState:
        """
        Returns the state for any mouse button.
        Parameters
        ----------
        button_id: One of MOUSE_BUTTON_n where n={1, .. , 8, LEFT, RIGHT, MIDDLE}
        Returns
        -------
        MouseButtonState
        """
        return self._buttons[button_id]

    def _poll_mouse(self, window):
        hover_state = glfw.get_window_attrib(window, glfw.HOVERED)
        self._hover = True if hover_state == 1 else False

        if self._hover:
            x, y = glfw.get_cursor_pos(window)
            self._dx = self._x - x
            self._dy = self._y - y
            self._x = x
            self._y = y

            for btn, btn_state in self._buttons.items():
                cur_state = glfw.get_mouse_button(window, btn)
                pressed = cur_state == glfw.PRESS
                released = cur_state == glfw.RELEASE

                if pressed and not btn_state.pressed and not btn_state.held:
                    btn_state.pressed = True
                    btn_state.released = False
                    btn_state.held = True
                elif pressed and btn_state.held:
                    btn_state.pressed = False
                    btn_state.released = False
                    btn_state.held = True
                elif released and btn_state.held:
                    btn_state.pressed = False
                    btn_state.released = True
                    btn_state.held = False
                else:
                    btn_state.pressed = False
                    btn_state.released = False
                    btn_state.held = False

    def _scroll_callback(self, window, x, y):
        self._scroll_dx += x
        self._scroll_dy += y



import freetype as ft
import numpy
from numpy.core.multiarray import ndarray


# This code is derived from the example in the *freetype-py* repository
# FreeType high-level python API - Copyright 2011-2015 Nicolas P. Rougier


class Font:
    def __init__(self, filename, size):
        self._font_data: ndarray = None
        self._glyph_width = None
        self._glyph_height = None
        self._grid_width = 16
        self._grid_height = 6
        self._make_font(filename, size)

    @property
    def glyph_width(self):
        return self._glyph_width

    @property
    def glyph_height(self):
        return self._glyph_height

    @property
    def grid_width(self):
        return self._grid_width

    @property
    def grid_height(self):
        return self._grid_height

    @property
    def data(self) -> ndarray:
        return self._font_data

    def _make_font(self, filename, size):
        face = ft.Face(filename)
        face.set_char_size(size * 64)
        if not face.is_fixed_width:
            raise UserWarning('Font is not monotype')

        grid_height = self.grid_height
        grid_width = self.grid_width
        # Determine largest glyph size
        width, height, ascender, descender = 0, 0, 0, 0
        # start at 32 to skip 32 first ASCII characters
        for c in range(32, grid_width * grid_height + 32):
            face.load_char(chr(c), ft.FT_LOAD_RENDER | ft.FT_LOAD_FORCE_AUTOHINT)
            bitmap = face.glyph.bitmap
            width = max(width, bitmap.width)
            ascender = max(ascender, face.glyph.bitmap_top)
            descender = max(descender, bitmap.rows - face.glyph.bitmap_top)
        height = ascender + descender

        # Generate texture data
        z = numpy.zeros((height * grid_height, width * grid_width), dtype=numpy.ubyte)
        for j in range(grid_height):
            for i in range(grid_width):
                pos = chr(32 + j * grid_width + i)
                face.load_char(pos, ft.FT_LOAD_RENDER | ft.FT_LOAD_FORCE_AUTOHINT)
                bitmap = face.glyph.bitmap
                x = i * width + face.glyph.bitmap_left
                y = j * height + ascender - face.glyph.bitmap_top
                z[y:y + bitmap.rows, x:x + bitmap.width].flat = bitmap.buffer
        self._font_data = z
        self._glyph_width = width
        self._glyph_height = height



import time

import glfw
import glm
from OpenGL.GL import GL_TRUE, glGetString, GL_VERSION, glViewport, glClearColor, \
    glClear, GL_COLOR_BUFFER_BIT, GL_DEPTH_BUFFER_BIT, glEnable, glBlendFunc, \
    GL_BLEND, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA

import pxng
import pxng.keys
import pxng.mouse
from pxng.grid import Grid
from pxng.colors import WHITE
from pxng.quad import Quad


class Window:

    def __init__(self, width, height, title='', **kwargs):
        """
        Creates a new window.

        After the successful creation of a window the loop has
        to be manually started with *start_event_loop*.
        To draw, use *set_update_handler* with a *fn(window)* callback.

        Parameters
        ----------
        width: int
            Width of window.
        height: int
            Height of window.
        title: str
            Title to show in title bar.
        **kwargs
            vsync: bool
                Turn off or on vsync. default=False
            resizable: bool
                Indicate if the windows should be resizable. default=False
            x_scale: float
                Scaling in 'x' direction. default=1
            y_scale: float
                Scaling in 'y' direction. default=1
            scale: float
                Sets 'x_scale' and 'y_scale' to the same value.
            hdpi: bool
                Use HDPI if available. Default=False
            color: tuple of float
                Set the background color. RGB or RGBA supported
        """
        if not glfw.init():
            raise UserWarning('Unable to initialize glfw')
        self.fps = 0
        self._handler = None
        self._title = title
        self.width = width
        self.height = height
        self.context = {}

        resizable = glfw.FALSE
        if 'resizable' in kwargs:
            resizable = glfw.TRUE if kwargs['resizable'] else glfw.FALSE
        glfw.window_hint(glfw.RESIZABLE, resizable)

        # switch to newer OpenGL version...
        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 3)
        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 3)
        glfw.window_hint(glfw.OPENGL_FORWARD_COMPAT, GL_TRUE)  # mac only
        glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)

        hdpi = glfw.FALSE
        if 'hdpi' in kwargs:
            hdpi = glfw.TRUE if kwargs['hdpi'] else glfw.FALSE
        glfw.window_hint(glfw.COCOA_RETINA_FRAMEBUFFER, hdpi)

        vsync = False
        if 'vsync' in kwargs:
            vsync = bool(kwargs['vsync'])

        self.x_scale = kwargs['x_scale'] if 'x_scale' in kwargs else 1
        self.y_scale = kwargs['y_scale'] if 'y_scale' in kwargs else 1
        if 'scale' in kwargs:
            self.x_scale = self.y_scale = kwargs['scale']

        self._window = glfw.create_window(width, height, title, None, None)

        self._color = (0, 0, 0, 1)
        if 'color' in kwargs:
            self._color = kwargs['color']

        if not self._window:
            glfw.terminate()
            raise UserWarning('Unable to create window')

        glfw.set_input_mode(self._window, glfw.STICKY_KEYS, glfw.TRUE)
        glfw.make_context_current(self._window)
        glfw.swap_interval(1 if vsync else 0)

        gl_version = glGetString(GL_VERSION).decode('utf-8')
        print(f'OpenGL Version string: {gl_version}')

        self._spaces = pxng.Spaces(self.width, self.height)
        self._spaces.projection.m = glm.ortho(0, width, height, 0, -1, 1)
        self._spaces.view.scale((self.x_scale, self.y_scale, 1))

        self._text_renderer = pxng.TextRenderer(self.create_default_font())
        self._elapsed_time = 0
        self._current_tint = WHITE
        self._key_poller = pxng.keys.KeyPoller()
        self._mouse_poller = pxng.mouse.Mouse(self._window)

        self._quad = Quad()
        self._grid = Grid(self.width, self.height)

    def create_default_font(self) -> pxng.Font:
        font_path = pxng.resource('fonts/C64_Pro_Mono-STYLE.ttf')
        return pxng.Font(font_path, 8)

    def start_event_loop(self):
        glViewport(0, 0, self.width, self.height)
        self._loop()

    def key_state(self, key) -> pxng.keys.KeyState:
        return self._key_poller.key_state(key)

    @property
    def mouse(self) -> pxng.mouse.Mouse:
        """
        Returns the mouse state.

        Returns
        -------
        pxng.mouse.Mouse
        """
        return self._mouse_poller

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title):
        self._title = title
        glfw.set_window_title(self._window, self._title)

    @property
    def elapsed_time(self) -> float:
        return self._elapsed_time

    def _loop(self):
        frame_count = 0
        fps_now = time.time()
        elapsed_now = time.time()
        while not glfw.window_should_close(self._window):
            self._key_poller.poll_keys(self._window)
            self._mouse_poller._poll_mouse(self._window)

            self._spaces.push()
            self._spaces.tint = WHITE

            glClearColor(*self._color)
            glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
            glEnable(GL_BLEND)
            glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

            now = time.time()
            self._elapsed_time = (now - elapsed_now)
            elapsed_now = now

            # draw call
            if self._handler is not None:
                self._handler(self)

            # render last batch of rects (if any)
            self._quad.draw_batch_if_started(self._spaces)

            # Swap front and back buffers
            glfw.swap_buffers(self._window)

            # Poll for and process events
            glfw.poll_events()

            frame_count += 1
            fps_time_delta = (time.time() - fps_now)
            self.fps = frame_count / fps_time_delta

            if fps_time_delta >= 1:
                title = f'{self.title} @ {self.fps:.0f} FPS'
                glfw.set_window_title(self._window, title)
                fps_now = time.time()
                frame_count = 0

        glfw.terminate()

    def close_window(self):
        glfw.set_window_should_close(self._window, glfw.TRUE)

    def set_update_handler(self, handler):
        self._handler = handler

    def draw_sprite(self, x, y, sprite, scale=1.0, tint=None):
        self._quad.draw_batch_if_started(self._spaces)
        self._spaces.tint = tint
        self._spaces.model.push()
        self._spaces.model.translate((x, y, 0))
        self._spaces.model.scale((scale, scale, 1))
        sprite.draw(self._spaces)
        self._spaces.model.pop()

    def draw_text(self, x, y, text, scale=1.0, tint=None, angle=0):
        self._quad.draw_batch_if_started(self._spaces)
        self._spaces.tint = tint
        self._text_renderer.draw_string(self._spaces, x, y, text, scale, angle)

    def draw_partial_sprite(self, x, y, sprite, sx, sy, sw, sh, scale=1.0, tint=None):
        self._quad.draw_batch_if_started(self._spaces)
        self._spaces.tint = tint
        self._spaces.model.push()
        self._spaces.model.translate((x, y, 0))
        self._spaces.model.scale((scale, scale, 1))
        sprite.draw_partial(self._spaces, sx, sy, sw, sh)
        self._spaces.model.pop()

    def draw_grid(self, size=10, tint=None, dash_size=4, gap_size=4):
        self._quad.draw_batch_if_started(self._spaces)
        self._spaces.tint = tint
        self._grid.draw(self._spaces, size, dash_size, gap_size)

    def fill_rect(self, x, y, w, h, tint=None):
        self._quad.check_started()
        self._spaces.tint = tint
        self._quad.draw(self._spaces, x, y, w, h)

    @property
    def tint(self):
        return self._spaces.default_tint

    @tint.setter
    def tint(self, tint):
        self._spaces.default_tint = tint



import glm
from OpenGL.GL import (GL_ARRAY_BUFFER, glGenBuffers, glBindBuffer, glBufferData,
                       GL_DYNAMIC_DRAW, glEnableVertexAttribArray, GL_UNSIGNED_INT,
                       GL_UNSIGNED_SHORT, GL_UNSIGNED_BYTE, GL_FLOAT,
                       glVertexAttribPointer, GL_ELEMENT_ARRAY_BUFFER, GL_DOUBLE)


class BufferObject:
    def __init__(self, data_type, array_type=GL_ARRAY_BUFFER, max_size=10000):
        self._data_type = data_type
        self._array_type = array_type
        self._arr = glm.array(data_type()) * max_size
        self._index = 0
        self._vbo = glGenBuffers(1)
        self._changed = True

    @property
    def type(self):
        return self._data_type

    @property
    def dtype(self):
        return self._arr.dtype

    @property
    def index(self):
        return self._index

    def __len__(self):
        return self._index

    @property
    def changed(self):
        return self._changed

    @property
    def component_count(self):
        return len(self._data_type())

    @property
    def bytes_per_element(self):
        return self._arr.dt_size * self.component_count

    def set_value(self, value):
        self._arr[self._index] = value
        self._index += 1
        self._changed = True

    def reset(self):
        self._index = 0
        self._changed = True

    def bind(self, attrib_index):
        if self._index == 0:
            return False

        glBindBuffer(self._array_type, self._vbo)

        if self._changed:
            data = self._arr
            size = data.itemsize * self._index
            glBufferData(self._array_type, size, data.ptr, GL_DYNAMIC_DRAW)
            self._changed = False

            if self._array_type == GL_ARRAY_BUFFER:
                glEnableVertexAttribArray(attrib_index)
                count = self.component_count
                stride = self.bytes_per_element
                dtype = self.dtype

                if dtype == 'float32':
                    gl_type = GL_FLOAT
                elif dtype == 'float64':
                    gl_type = GL_DOUBLE
                elif dtype == 'uint8':
                    gl_type = GL_UNSIGNED_BYTE
                elif dtype == 'uint16':
                    gl_type = GL_UNSIGNED_SHORT
                elif dtype == 'uint32':
                    gl_type = GL_UNSIGNED_INT
                else:
                    raise UserWarning(f'Unknown data type: {dtype}')

                glVertexAttribPointer(attrib_index, count, gl_type, False, stride, None)

        return True



from ._utils import resource
from . import colors
from .spaces import Spaces
from .font import Font
from .sprite import Sprite
from .animated_sprite import AnimatedSprite
from .text_renderer import TextRenderer
from .shader import ShaderProgram, ShaderType
from .buffer_object import BufferObject
from .vertex_array_object import VertexArrayObject
from .window import Window





from typing import Tuple

import glm
import imageio
from numpy.core.multiarray import ndarray

import pxng
from pxng import resource
from OpenGL.GL import GL_TRIANGLES, GL_RGBA, GL_RGB, glGenTextures, \
    GL_TEXTURE_RECTANGLE, glBindTexture, glTexParameteri, GL_TEXTURE_MAG_FILTER, \
    GL_NEAREST, GL_TEXTURE_MIN_FILTER, glTexImage2D, GL_UNSIGNED_BYTE, glTexSubImage2D, \
    GL_BLEND, glDisable, GL_RED, glPixelStorei, GL_UNPACK_ALIGNMENT, glGetInteger


class SpriteRectangle:
    def __init__(self):
        program = pxng.ShaderProgram('SpriteShader')
        program.add_shader(resource('shaders/sprite.vert'), pxng.ShaderType.Vertex)
        program.add_shader(resource('shaders/sprite.frag'), pxng.ShaderType.Fragment)
        program.compile_and_link()

        program.add_uniform('projection_view', glm.mat4x4)
        program.add_uniform('model', glm.mat4x4)
        program.add_uniform('texture_matrix', glm.mat4x4)
        program.add_uniform('color', glm.vec4)
        program.add_uniform('sprite_texture', glm.ivec1)
        self._program = program

        self._vao = pxng.VertexArrayObject(GL_TRIANGLES)
        self._vao.attach_buffer(pxng.BufferObject(data_type=glm.vec3))  # vertex buffer
        self._vao.attach_buffer(pxng.BufferObject(data_type=glm.uvec2))  # texture buffer

        self._vao.add_quad(
            glm.vec3(0, 0, 0),
            glm.vec3(0, -1, 0),
            glm.vec3(1, -1, 0),
            glm.vec3(1, 0, 0),
        )

        self._vao.set_texture(
            glm.uvec2(0, 1),
            glm.uvec2(0, 0),
            glm.uvec2(1, 0),
            glm.uvec2(1, 1)
        )

    def draw(self, spaces: pxng.Spaces):
        if self._vao.bind():
            self._program.activate()
            self._program.set_uniform('projection_view', spaces.projection_view)
            self._program.set_uniform('model', spaces.model.m)
            self._program.set_uniform('texture_matrix', spaces.texture.m)
            self._program.set_uniform('color', spaces.tint)
            self._program.set_uniform('sprite_texture', 0)
            self._vao.draw()


class Sprite:
    def __init__(self, data: ndarray):
        self._data = data
        self._created = False
        self._dirty = False
        self._texid = None
        self._width = data.shape[1]
        self._height = data.shape[0]
        self._components = 1
        if len(data.shape) > 2:
            self._components = data.shape[2]

        if self._components == 4:
            self._format = GL_RGBA
        elif self._components == 3:
            self._format = GL_RGB
        elif self._components == 1:
            self._format = GL_RED

        self._rect = None

    @property
    def width(self):
        return self._width

    @property
    def height(self):
        return self._height

    def set_pixel(self, x, y, color: Tuple[int, int, int]):
        self._data[y, x] = color
        self._dirty = True

    def _set_unpack_alignment(self):
        self._unpack_alignment = glGetInteger(GL_UNPACK_ALIGNMENT)
        if self._width % 2 == 1 or self._width % 4 == 2:
            # Odd sized texture
            glPixelStorei(GL_UNPACK_ALIGNMENT, 1)

    def _reset_unpack_alignment(self):
        glPixelStorei(GL_UNPACK_ALIGNMENT, self._unpack_alignment)

    def _create(self):
        self._texid = glGenTextures(1)
        glBindTexture(GL_TEXTURE_RECTANGLE, self._texid)

        glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST)

        d = self._data
        w = self._width
        h = self._height
        fmt = self._format

        self._set_unpack_alignment()
        glTexImage2D(GL_TEXTURE_RECTANGLE, 0, fmt, w, h, 0, fmt, GL_UNSIGNED_BYTE, d)
        self._reset_unpack_alignment()

        self._rect = SpriteRectangle()
        self._created = True

    def _update(self):
        glBindTexture(GL_TEXTURE_RECTANGLE, self._texid)

        d = self._data
        w = self._width
        h = self._height
        fmt = self._format

        self._set_unpack_alignment()
        glTexSubImage2D(GL_TEXTURE_RECTANGLE, 0, 0, 0, w, h, fmt, GL_UNSIGNED_BYTE, d)
        self._reset_unpack_alignment()

        self._updated = True

    def update(self):
        self._dirty = True

    def activate(self):
        if not self._created:
            self._create()

        if self._dirty:
            self._update()

        glBindTexture(GL_TEXTURE_RECTANGLE, self._texid)
        glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_RECTANGLE, GL_TEXTURE_MIN_FILTER, GL_NEAREST)

    def deactivate(self):
        glBindTexture(GL_TEXTURE_RECTANGLE, 0)

    def _pre_draw(self, spaces: pxng.Spaces):
        self.activate()
        spaces.model.push()
        spaces.texture.push()

    def _post_draw(self, spaces: pxng.Spaces):
        self.deactivate()
        spaces.model.pop()
        spaces.texture.pop()

    def draw(self, spaces: pxng.Spaces):
        """
        Draw the sprite at the current position.

        Parameters
        ----------
        spaces : pxng.Spaces
            the current coordinate systems
        """
        self._pre_draw(spaces)
        spaces.model.translate((0, self._height, 0))  # anchor is upper left
        spaces.model.scale((self._width, self._height, 1))
        spaces.texture.scale(glm.vec3(self._width, self._height, 1))
        self._rect.draw(spaces)
        self._post_draw(spaces)

    def draw_partial(self, spaces: pxng.Spaces, x, y, width, height):
        """
        Draw a partial sprite using pixel coordinates for the partial image data.
        The coordinates represent a sub region in the sprite.

        Parameters
        ----------
        spaces: pxng.Spaces
            the current coordinate systems
        x: int
            x coordinate in pixel space
        y: int
            y coordinate in pixel space
        width: int
            width in number of pixels
        height: int
            height in number of pixels
        """
        self._pre_draw(spaces)
        spaces.model.translate((0, height, 0))  # anchor is upper left
        spaces.model.scale((width, height, 1))

        spaces.texture.translate(glm.vec3(x, y, 0))
        spaces.texture.scale(glm.vec3(width, height, 1))
        self._rect.draw(spaces)

        self._post_draw(spaces)

    @classmethod
    def create_from_image(cls, path):
        img_data = imageio.imread(path)
        return Sprite(img_data)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pxng",
    version="0.0.13",
    author="Jean-Paul Balabanian",
    author_email="jepebe@prador.net",
    description="A library for fiddling with pixels",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jepebe/pixelengine",
    packages=setuptools.find_packages(),
    license='BSD',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'glfw>=2.0.0',
        'pyopengl>=3.1.0',
        'freetype-py>=2.2.0',
        'numpy>=1.19.0',
        'imageio>=2.9.0',
        'pyglm>=1.99.0',
    ],
    extras_require={
        ':python_version < "3.7"': [
            'dataclasses',
        ],
    },
    package_data={'pxng': [
        'resources/fonts/C64_Pro_Mono-STYLE.ttf',
        'resources/shaders/*',
    ]
    },
    python_requires='>=3.6',
)



from typing import Tuple, List

import pxng


class AnimatedSprite:
    def __init__(self, sprite: pxng.Sprite, grid_size: Tuple[int, int]):
        self._sprite = sprite
        self._grid_width = grid_size[0]
        self._grid_height = grid_size[1]

        self._animations = {}
        all_frames = self._create_frames(self._grid_width, self._grid_height)
        self.set_animation('default', all_frames, 30)
        self.set_current_animation('default')

        self._sw = self._sprite.width // self._grid_width
        self._sh = self._sprite.height // self._grid_height

    def _create_frames(self, w, h):
        return [(i % w, i // w) for i in range(w * h)]

    def set_animation(self, name: str, frames: List[Tuple[int, int]], fps: int):
        self._animations[name] = {
            'name': name,
            'frames': frames,
            'fps': fps,
            'time_delta': 1 / fps,
            'accumulated_time': 0,
            'current_frame': 0,
            'frame_count': len(frames)
        }

    def set_current_animation(self, name):
        self._current_animation = self._animations[name]

    def advance_time_all(self, elapsed_time):
        for item in self._animations.values():
            item['accumulated_time'] += elapsed_time

            if item['accumulated_time'] >= item['time_delta']:
                item['accumulated_time'] = 0
                item['current_frame'] += 1

            if item['current_frame'] >= item['frame_count']:
                item['current_frame'] = 0

    def advance_time(self, elapsed_time):
        item = self._current_animation
        item['accumulated_time'] += elapsed_time

        if item['accumulated_time'] >= item['time_delta']:
            item['accumulated_time'] = 0
            item['current_frame'] += 1

        if item['current_frame'] >= item['frame_count']:
            item['current_frame'] = 0

    def draw(self, spaces: pxng.Spaces):
        animation = self._current_animation
        current_frame = animation['current_frame']
        x, y = animation['frames'][current_frame]

        w = self._sw
        h = self._sh

        self._sprite.draw_partial(spaces, x * w, y * h, w, h)





from dataclasses import dataclass
from typing import Dict

import glfw

KEY_SPACE = glfw.KEY_SPACE
KEY_APOSTROPHE = glfw.KEY_APOSTROPHE
KEY_COMMA = glfw.KEY_COMMA
KEY_MINUS = glfw.KEY_MINUS
KEY_PERIOD = glfw.KEY_PERIOD
KEY_SLASH = glfw.KEY_SLASH
KEY_0 = glfw.KEY_0
KEY_1 = glfw.KEY_1
KEY_2 = glfw.KEY_2
KEY_3 = glfw.KEY_3
KEY_4 = glfw.KEY_4
KEY_5 = glfw.KEY_5
KEY_6 = glfw.KEY_6
KEY_7 = glfw.KEY_7
KEY_8 = glfw.KEY_8
KEY_9 = glfw.KEY_9
KEY_SEMICOLON = glfw.KEY_SEMICOLON
KEY_EQUAL = glfw.KEY_EQUAL
KEY_A = glfw.KEY_A
KEY_B = glfw.KEY_B
KEY_C = glfw.KEY_C
KEY_D = glfw.KEY_D
KEY_E = glfw.KEY_E
KEY_F = glfw.KEY_F
KEY_G = glfw.KEY_G
KEY_H = glfw.KEY_H
KEY_I = glfw.KEY_I
KEY_J = glfw.KEY_J
KEY_K = glfw.KEY_K
KEY_L = glfw.KEY_L
KEY_M = glfw.KEY_M
KEY_N = glfw.KEY_N
KEY_O = glfw.KEY_O
KEY_P = glfw.KEY_P
KEY_Q = glfw.KEY_Q
KEY_R = glfw.KEY_R
KEY_S = glfw.KEY_S
KEY_T = glfw.KEY_T
KEY_U = glfw.KEY_U
KEY_V = glfw.KEY_V
KEY_W = glfw.KEY_W
KEY_X = glfw.KEY_X
KEY_Y = glfw.KEY_Y
KEY_Z = glfw.KEY_Z
KEY_LEFT_BRACKET = glfw.KEY_LEFT_BRACKET
KEY_BACKSLASH = glfw.KEY_BACKSLASH
KEY_RIGHT_BRACKET = glfw.KEY_RIGHT_BRACKET
KEY_GRAVE_ACCENT = glfw.KEY_GRAVE_ACCENT
KEY_WORLD_1 = glfw.KEY_WORLD_1
KEY_WORLD_2 = glfw.KEY_WORLD_2
KEY_ESCAPE = glfw.KEY_ESCAPE
KEY_ENTER = glfw.KEY_ENTER
KEY_TAB = glfw.KEY_TAB
KEY_BACKSPACE = glfw.KEY_BACKSPACE
KEY_INSERT = glfw.KEY_INSERT
KEY_DELETE = glfw.KEY_DELETE
KEY_RIGHT = glfw.KEY_RIGHT
KEY_LEFT = glfw.KEY_LEFT
KEY_DOWN = glfw.KEY_DOWN
KEY_UP = glfw.KEY_UP
KEY_PAGE_UP = glfw.KEY_PAGE_UP
KEY_PAGE_DOWN = glfw.KEY_PAGE_DOWN
KEY_HOME = glfw.KEY_HOME
KEY_END = glfw.KEY_END
KEY_CAPS_LOCK = glfw.KEY_CAPS_LOCK
KEY_SCROLL_LOCK = glfw.KEY_SCROLL_LOCK
KEY_NUM_LOCK = glfw.KEY_NUM_LOCK
KEY_PRINT_SCREEN = glfw.KEY_PRINT_SCREEN
KEY_PAUSE = glfw.KEY_PAUSE
KEY_F1 = glfw.KEY_F1
KEY_F2 = glfw.KEY_F2
KEY_F3 = glfw.KEY_F3
KEY_F4 = glfw.KEY_F4
KEY_F5 = glfw.KEY_F5
KEY_F6 = glfw.KEY_F6
KEY_F7 = glfw.KEY_F7
KEY_F8 = glfw.KEY_F8
KEY_F9 = glfw.KEY_F9
KEY_F10 = glfw.KEY_F10
KEY_F11 = glfw.KEY_F11
KEY_F12 = glfw.KEY_F12
KEY_F13 = glfw.KEY_F13
KEY_F14 = glfw.KEY_F14
KEY_F15 = glfw.KEY_F15
KEY_F16 = glfw.KEY_F16
KEY_F17 = glfw.KEY_F17
KEY_F18 = glfw.KEY_F18
KEY_F19 = glfw.KEY_F19
KEY_F20 = glfw.KEY_F20
KEY_F21 = glfw.KEY_F21
KEY_F22 = glfw.KEY_F22
KEY_F23 = glfw.KEY_F23
KEY_F24 = glfw.KEY_F24
KEY_F25 = glfw.KEY_F25
KEY_KP_0 = glfw.KEY_KP_0
KEY_KP_1 = glfw.KEY_KP_1
KEY_KP_2 = glfw.KEY_KP_2
KEY_KP_3 = glfw.KEY_KP_3
KEY_KP_4 = glfw.KEY_KP_4
KEY_KP_5 = glfw.KEY_KP_5
KEY_KP_6 = glfw.KEY_KP_6
KEY_KP_7 = glfw.KEY_KP_7
KEY_KP_8 = glfw.KEY_KP_8
KEY_KP_9 = glfw.KEY_KP_9
KEY_KP_DECIMAL = glfw.KEY_KP_DECIMAL
KEY_KP_DIVIDE = glfw.KEY_KP_DIVIDE
KEY_KP_MULTIPLY = glfw.KEY_KP_MULTIPLY
KEY_KP_SUBTRACT = glfw.KEY_KP_SUBTRACT
KEY_KP_ADD = glfw.KEY_KP_ADD
KEY_KP_ENTER = glfw.KEY_KP_ENTER
KEY_KP_EQUAL = glfw.KEY_KP_EQUAL
KEY_LEFT_SHIFT = glfw.KEY_LEFT_SHIFT
KEY_LEFT_CONTROL = glfw.KEY_LEFT_CONTROL
KEY_LEFT_ALT = glfw.KEY_LEFT_ALT
KEY_LEFT_SUPER = glfw.KEY_LEFT_SUPER
KEY_RIGHT_SHIFT = glfw.KEY_RIGHT_SHIFT
KEY_RIGHT_CONTROL = glfw.KEY_RIGHT_CONTROL
KEY_RIGHT_ALT = glfw.KEY_RIGHT_ALT
KEY_RIGHT_SUPER = glfw.KEY_RIGHT_SUPER
KEY_MENU = glfw.KEY_MENU

_keys = [KEY_SPACE, KEY_APOSTROPHE, KEY_COMMA, KEY_MINUS, KEY_PERIOD,
         KEY_SLASH, KEY_0, KEY_1, KEY_2, KEY_3, KEY_4, KEY_5, KEY_6, KEY_7, KEY_8,
         KEY_9, KEY_SEMICOLON, KEY_EQUAL, KEY_A, KEY_B, KEY_C, KEY_D, KEY_E, KEY_F,
         KEY_G, KEY_H, KEY_I, KEY_J, KEY_K, KEY_L, KEY_M, KEY_N, KEY_O, KEY_P, KEY_Q,
         KEY_R, KEY_S, KEY_T, KEY_U, KEY_V, KEY_W, KEY_X, KEY_Y, KEY_Z,
         KEY_LEFT_BRACKET, KEY_BACKSLASH, KEY_RIGHT_BRACKET, KEY_GRAVE_ACCENT,
         KEY_WORLD_1, KEY_WORLD_2, KEY_ESCAPE, KEY_ENTER, KEY_TAB, KEY_BACKSPACE,
         KEY_INSERT, KEY_DELETE, KEY_RIGHT, KEY_LEFT, KEY_DOWN, KEY_UP, KEY_PAGE_UP,
         KEY_PAGE_DOWN, KEY_HOME, KEY_END, KEY_CAPS_LOCK, KEY_SCROLL_LOCK, KEY_NUM_LOCK,
         KEY_PRINT_SCREEN, KEY_PAUSE, KEY_F1, KEY_F2, KEY_F3, KEY_F4, KEY_F5, KEY_F6,
         KEY_F7, KEY_F8, KEY_F9, KEY_F10, KEY_F11, KEY_F12, KEY_F13, KEY_F14, KEY_F15,
         KEY_F16, KEY_F17, KEY_F18, KEY_F19, KEY_F20, KEY_F21, KEY_F22, KEY_F23,
         KEY_F24, KEY_F25, KEY_KP_0, KEY_KP_1, KEY_KP_2, KEY_KP_3, KEY_KP_4, KEY_KP_5,
         KEY_KP_6, KEY_KP_7, KEY_KP_8, KEY_KP_9, KEY_KP_DECIMAL, KEY_KP_DIVIDE,
         KEY_KP_MULTIPLY, KEY_KP_SUBTRACT, KEY_KP_ADD, KEY_KP_ENTER, KEY_KP_EQUAL,
         KEY_LEFT_SHIFT, KEY_LEFT_CONTROL, KEY_LEFT_ALT, KEY_LEFT_SUPER,
         KEY_RIGHT_SHIFT, KEY_RIGHT_CONTROL, KEY_RIGHT_ALT, KEY_RIGHT_SUPER, KEY_MENU]


@dataclass
class KeyState:
    pressed: bool = False
    released: bool = False
    held: bool = False

    def __str__(self):
        return f'self.pressed={self.pressed} self.released={self.released} self.held={self.held}'


class KeyPoller:
    def __init__(self):
        self._key_states: Dict[int, KeyState] = {key: KeyState() for key in _keys}

    def poll_keys(self, window):
        for key, key_state in self._key_states.items():
            cur_state = glfw.get_key(window, key)
            pressed = cur_state == glfw.PRESS
            released = cur_state == glfw.RELEASE

            if pressed and not key_state.pressed and not key_state.held:
                key_state.pressed = True
                key_state.released = False
                key_state.held = True
            elif pressed and key_state.held:
                key_state.pressed = False
                key_state.released = False
                key_state.held = True
            elif released and key_state.held:
                key_state.pressed = False
                key_state.released = True
                key_state.held = False
            else:
                key_state.pressed = False
                key_state.released = False
                key_state.held = False

    def key_state(self, key) -> KeyState:
        return self._key_states[key]



import glm
from OpenGL.GL import GL_TRIANGLES

import pxng
from pxng import resource


class Quad:
    def __init__(self):
        program = pxng.ShaderProgram('QuadShader')
        program.add_shader(resource('shaders/quad.vert'), pxng.ShaderType.Vertex)
        program.add_shader(resource('shaders/quad.frag'), pxng.ShaderType.Fragment)
        program.compile_and_link()

        program.add_uniform('projection_view', glm.mat4x4)

        self._shader_program = program

        self._vao = pxng.VertexArrayObject(GL_TRIANGLES)
        self._vao.attach_buffer(pxng.BufferObject(data_type=glm.vec3))  # vertex buffer
        self._vao.attach_buffer(pxng.BufferObject(data_type=glm.vec4))  # color buffer

        self._started = False

    def start_batch(self):
        self._vao.reset()
        self._started = True

    def check_started(self):
        if not self._started:
            self.start_batch()

    def draw(self, spaces: pxng.Spaces, x, y, w, h):
        p1 = glm.vec3(x, y, 0)
        p2 = glm.vec3(x, y + h, 0)
        p3 = glm.vec3(x + w, y + h, 0)
        p4 = glm.vec3(x + w, y, 0)

        self._vao.add_quad(p1, p2, p3, p4)
        tint = spaces.tint
        self._vao.set_colors(tint, tint, tint, tint)

    def draw_batch(self, spaces: pxng.Spaces):
        if self._vao.bind():
            program = self._shader_program
            program.activate()
            program.set_uniform('projection_view', spaces.projection_view)
            self._vao.draw()
            self._started = False

    def draw_batch_if_started(self, spaces):
        if self._started:
            self.draw_batch(spaces)



import glm

from OpenGL.GL import GL_ARRAY_BUFFER, glGenVertexArrays, glBindVertexArray, \
    glDrawArrays, GL_POINTS

import pxng


class FontRenderer:
    def __init__(self):
        self._vao = glGenVertexArrays(1)
        self._points = pxng.BufferObject(data_type=glm.vec3, array_type=GL_ARRAY_BUFFER)
        self._chars = pxng.BufferObject(data_type=glm.uvec1, array_type=GL_ARRAY_BUFFER)

        program = pxng.ShaderProgram('FontShader')
        program.add_shader(pxng.resource('shaders/font.vert'), pxng.ShaderType.Vertex)
        program.add_shader(pxng.resource('shaders/font.geom'), pxng.ShaderType.Geometry)
        program.add_shader(pxng.resource('shaders/font.frag'), pxng.ShaderType.Fragment)
        program.compile_and_link()

        program.add_uniform('projection_view', glm.mat4x4)
        program.add_uniform('model', glm.mat4x4)
        program.add_uniform('color', glm.vec4)
        program.add_uniform('size', glm.uvec1)
        program.add_uniform('sprite_texture', glm.ivec1)
        self._program = program

    def add_char(self, pos: glm.vec3, char: int):
        self._points.set_value(pos)
        self._chars.set_value(glm.uvec1(char))

    def reset(self):
        self._points.reset()
        self._chars.reset()

    def draw(self, spaces):
        glBindVertexArray(self._vao)
        if self._points.bind(0) and self._chars.bind(1):
            self._program.activate()
            self._program.set_uniform('projection_view', spaces.projection_view)
            self._program.set_uniform('model', spaces.model.m)
            self._program.set_uniform('color', spaces.tint)
            # self._program.set_uniform('size', 8)
            self._program.set_uniform('sprite_texture', 0)

            glDrawArrays(GL_POINTS, 0, len(self._points))

        glBindVertexArray(0)


class TextRenderer:
    def __init__(self, font: pxng.Font):
        self._font = font
        self._font_sprite = pxng.Sprite(self._font._font_data)
        self._renderer = FontRenderer()

    def draw_string(self, spaces: pxng.Spaces, x, y, text, scale=1.0, angle=0):
        self._renderer.reset()
        self._font_sprite.activate()

        angle = glm.radians(angle)

        dx = self._font.glyph_width
        for i, c in enumerate(text):
            self._renderer.add_char(glm.vec3(i * dx, 0, 0), ord(c))

        spaces.model.push()
        spaces.model.translate((x, y, 1))
        spaces.model.scale((scale, scale, 1))
        spaces.model.rotate(angle, (0, 0, 1))

        self._renderer.draw(spaces)

        spaces.model.pop()
        self._font_sprite.deactivate()
